dzlsuggestionpopover {
  background: transparent;
}

dzlsuggestionpopover > revealer > box {
  margin: 0px 0px 60px 0px;
  border-radius: 0px;
  border-width: 0px;
  border-bottom: 1px solid @borders;
  box-shadow: 0px -6px 12px #000;
}

dzlsuggestionpopover > revealer > box > elastic > scrolledwindow {
  background: alpha(@content_view_bg, 0.97);
  border-radius: 0 0 7px 7px;
}

dzlsuggestionpopover > revealer > box > elastic > scrolledwindow > viewport > list {
  background: transparent;
}

dzlsuggestionpopover > revealer > box > elastic > scrolledwindow > viewport > list > row:first-child {
  transition: none;
}

dzlsuggestionpopover > revealer > box > elastic > scrolledwindow > viewport > list > row {
  transition: none;
  margin-left: 100px;
  margin-right: 100px;
}

dzlsuggestionpopover > revealer > box > elastic > scrolledwindow > viewport > list > row:selected .title {
  color: @theme_selected_fg_color;
}

dzlsuggestionpopover > revealer > box > elastic > scrolledwindow > viewport > list > row .title {
  color: @theme_selected_bg_color;
}

dzlsuggestionpopover > revealer > box > elastic > scrolledwindow > viewport > list > row:first-child:selected {
  border-radius: 0;
}

dzlsuggestionpopover > revealer > box > elastic > scrolledwindow > viewport > list > row > box {
  margin: 4px 8px 4px 8px;
}

dzlsuggestionpopover > revealer > box > elastic > scrolledwindow > viewport > list > row:last-child {
  border-bottom: none;
}
